// $Id: MultiPageDocumentViewport.java 16840 2013-01-28 17:01:39Z mike $

package org.faceless.pdf2.viewer2;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.beans.*;
import java.util.List;
import javax.swing.*;
import org.faceless.pdf2.*;

/**
 * A {@link DocumentViewport} that displays all the pages of a PDF in one column
 * (or row) in the viewport. Navigating between pages is done by scrolling the viewport.
 *
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @see DocumentPanel
 * @since 2.10.4
 */
public class MultiPageDocumentViewport extends DocumentViewport {

    private JScrollPane scrollPane;
    private View view;
    private float zoom;

    private Listener listener;
    private RenderingHints hints;
    private ArrayList pagePanelListeners;
    private ArrayList pagePanelInteractionListeners;
    private PDF pdf;
    private PDFParser parser;

    /**
     * Constructs a new MultiPageDocumentViewport with a vertical
     * orientation.
     */
    public MultiPageDocumentViewport() {
        this(Adjustable.VERTICAL);
    }

    /**
     * Constructs a new multi-page document viewport with the specified
     * orientation.
     * @param orientation Adjustable.VERTICAL or Adjustable.HORIZONTAL
     */
    public MultiPageDocumentViewport(int orientation) {
        view = this.new View(orientation);
        setRenderingHints(getDefaultRenderingHints());
        setLayout(new BorderLayout());
        scrollPane = new JScrollPane(view) {
            protected JViewport createViewport() {
                return new JViewport() {
                    public void scrollRectToVisible(Rectangle r) {
                        // NB eat this event to prevent caret from scrolling viewport
                    }
                };
            }
        };
        add(scrollPane, BorderLayout.CENTER);
        zoom = 1;

        // event management
        listener = this.new Listener();
        addComponentListener(listener);
        scrollPane.getHorizontalScrollBar().addAdjustmentListener(listener);
        scrollPane.getVerticalScrollBar().addAdjustmentListener(listener);

        pagePanelListeners = new ArrayList();
        pagePanelInteractionListeners = new ArrayList();

        Util.fixScrollPaneKeyBindings(scrollPane);
    }

    public void setDocumentPanel(DocumentPanel panel) {
        super.setDocumentPanel(panel);
        if (panel != null) {
            setPDF(panel.getPDF());
            parser = panel.getParser();
            view.reorder();
            scrollPane.getHorizontalScrollBar().setUnitIncrement(mouseWheelUnit);
            scrollPane.getVerticalScrollBar().setUnitIncrement(mouseWheelUnit);
        } else {
            pdf = null;
            parser = null;
            view.dispose();
        }
    }

    public boolean isDraggable() {
        return view.isDraggable();
    }

    public void setDraggable(boolean draggable) {
        view.setDraggable(draggable);
    }

    public void setPage(PDFPage page, double x, double y, double zoom) {
        setPDF(page.getPDF());
        setZoom((float) zoom);
        Rectangle2D rect = PagePanel.getFullPageView(page);
        ensureVisible(page, x, rect.getHeight() - y, false);
    }

    public void ensureVisible(PDFPage page, double x, double y) {
        ensureVisible(page, x, y, true);
    }

    /**
     * @param x x coordinate in PDF units
     * @param y y coordinate in PDF units
     * @param centre if true, centre this location in the viewport,
     * otherwise cause it to be displayed at the top left of the viewport.
     */
    private void ensureVisible(PDFPage page, double x, double y, boolean centre) {
        Rectangle pageRect = view.getPageRectangle(page.getPageNumber() - 1);
        if (pageRect != null) {
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            int xoffset = 0, yoffset = 0;
            if (!Double.isNaN(x)) {
                xoffset = pointsToPixels((float) x);
            }
            if (!Double.isNaN(y)) {
                yoffset = pointsToPixels((float) y);
                // convert to AWT axis space
                yoffset = pageRect.height - yoffset;
            }
            xoffset = pageRect.x + xoffset;
            yoffset = pageRect.y + yoffset;
            if (centre) {
                xoffset -= hsb.getVisibleAmount() / 2;
                yoffset -= vsb.getVisibleAmount() / 2;
            }
            if (getPagePanel() != null && getPagePanel().getClip() != null) {
                hsb.setValue(xoffset);  // Don't want to smooth scroll if we're clipping
                vsb.setValue(yoffset);  // the page rectangles.
            } else {
                smoothScroll(xoffset, yoffset, hsb, vsb);
            }
        }
    }

    public boolean isPageVisible(PDFPage page, double x, double y) {
        Rectangle pageRect = view.getPageRectangle(page.getPageNumber() - 1);
        if (pageRect != null) {
            Rectangle target = (Rectangle) pageRect.clone();
            float dpi = getZoom() * Util.getScreenResolution(this);
            if (!Double.isNaN(x)) {
                int xoffset = Math.round((float) x * (dpi / 72));
                target.x += xoffset;
                target.width = 8;
            }
            if (!Double.isNaN(y)) {
                int yoffset = Math.round((float) y * (dpi / 72));
                // convert to PDF axis space
                yoffset = pageRect.height - yoffset;
                target.y += yoffset;
                target.height = 8;
            }
            return getClip().intersects(target);
        }
        return false;
    }

    public JComponent getView() {
        return (JComponent)scrollPane.getViewport().getView();
    }

    public float getZoom() {
        if (zoom==0) {
            zoom = getTargetZoom(getZoomMode(), getPage());
        }
        return zoom;
    }

    public void setZoom(float zoom) {
        if (!SwingUtilities.isEventDispatchThread()) {
            throw new IllegalStateException("Must be on EDT");
        }
        if (zoom == zoom && zoom > 0.01 && Math.abs(zoom - this.zoom) > 0.01) {
            final JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            final JScrollBar vsb = scrollPane.getVerticalScrollBar();
            final int hx = hsb.getVisibleAmount();
            final int vx = vsb.getVisibleAmount();
            final float hv = (float)(hsb.getValue() + hx/2) / hsb.getMaximum();
            final float vv = (float)(vsb.getValue() + vx/2) / vsb.getMaximum();
            this.zoom = zoom;
            view.redimension();

            final Dimension viewsize = view.getSize();
            listener.setEnabled(false); // don't send adjustment events
            hsb.setValues(Math.round(viewsize.width * hv) - hx/2, hx, 0, viewsize.width);
            vsb.setValues(Math.round(viewsize.height * vv) - vx/2, vx, 0, viewsize.height);
            listener.setEnabled(true);
            view.revalidate();
            view.repaint();
        }
    }

    private int pointsToPixels(float points) {
        return Math.round(points * (getZoom() * Util.getScreenResolution(this) / 72));
    }

    private float pixelsToPoints(int pixels) {
        return (float)pixels / (getZoom() * Util.getScreenResolution(this) / 72);
    }

    public Dimension getViewportSize() {
        return scrollPane.getViewport().getSize();
    }

    public PagePanel getPagePanel() {
        return view.getPagePanel();
    }

    public Collection getPagePanels() {
        return view.getPagePanels();
    }

    public PDFPage getRenderingPage() {
        return getPage();
    }

    private void setPDF(PDF pdf) {
        if (this.pdf != pdf) {    // Only happens once unless not part of a DocumentPanel
            this.pdf = null;
            view.reorder();
            this.pdf = pdf;
            view.reorder();
            view.redimension();
        }
    }

    private Rectangle getClip() {
        JScrollBar hsb = scrollPane.getHorizontalScrollBar();
        JScrollBar vsb = scrollPane.getVerticalScrollBar();
        return new Rectangle(hsb.getValue(), vsb.getValue(), hsb.getVisibleAmount(), vsb.getVisibleAmount());
    }

    public Adjustable getAdjustable(int position) {
        switch (position) {
          case Adjustable.HORIZONTAL:
            return scrollPane.getHorizontalScrollBar();
          case Adjustable.VERTICAL:
            return scrollPane.getVerticalScrollBar();
          default:
            throw new IllegalArgumentException(Integer.toString(position));
        }
    }

    public void setAdjustableValues(int horizontal, int vertical) {
        scrollPane.getHorizontalScrollBar().setValue(horizontal);
        scrollPane.getVerticalScrollBar().setValue(vertical);
        view.revalidate();
        view.repaint();
    }

    public void setRenderingHints(RenderingHints hints) {
        this.hints = hints;
        for (Iterator i = view.getPagePanels().iterator();i.hasNext();) {
            ((PagePanel)i.next()).setRenderingHints(hints);
        }
    }

    public void addPagePanelListener(PagePanelListener listener) {
        synchronized (pagePanelListeners) {
            pagePanelListeners.add(listener);
        }
    }

    public void removePagePanelListener(PagePanelListener listener) {
        synchronized (pagePanelListeners) {
            pagePanelListeners.remove(listener);
        }
    }

    public void addPagePanelInteractionListener(PagePanelInteractionListener listener) {
        synchronized (pagePanelInteractionListeners) {
            pagePanelInteractionListeners.add(listener);
        }
    }

    public void removePagePanelInteractionListener(PagePanelInteractionListener listener) {
        synchronized (pagePanelInteractionListeners) {
            pagePanelInteractionListeners.remove(listener);
        }
    }

    public void setZoomMode(int zoommode) {
        super.setZoomMode(zoommode);
        if (getPage() != null) {
            setZoom(getTargetZoom(zoommode, getPage()));
        }
    }

    public float getTargetZoom(int zoommode, PDFPage page) {
        float outzoom = zoom;
        if (page!=null) {
            Dimension pps = view.getLargestPageSize();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            Dimension size = new Dimension(hsb.getVisibleAmount(), vsb.getVisibleAmount());
            switch (zoommode) {
              case ZOOM_FITWIDTH:
                size.width -= vsb.getWidth();
                outzoom = (float)size.width / pps.width;
                break;
              case ZOOM_FITHEIGHT:
                size.height -= hsb.getHeight();
                outzoom = (float)size.height / pps.height;
                break;
              case ZOOM_FIT:
                size.width -= vsb.getWidth();
                size.height -= hsb.getHeight();
                float zw = (float)size.width / pps.width;
                float zh = (float)size.height / pps.height;
                outzoom = zw < zh ? zw : zh;
                break;
              default:
                outzoom = zoom==0 ? 1 : zoom;
            }
        }
        return outzoom;
    }

    public void propertyChange(PropertyChangeEvent event) {
        String name = event.getPropertyName();
        Object source = event.getSource();
        boolean relayout = false;
        if (name.equals("pages") && source == pdf) {
            relayout = true;
        } else if ((name.endsWith("Box") || name.equals("orientation")) && source instanceof PDFPage) {
            relayout = true;
        }
        if (relayout) {
            smoothScroll(0, 0, null, null);
            view.reorder();
            view.redimension();
            view.revalidate();
            view.repaint();
        }
    }

    private class View extends JPanel implements PagePanelListener, PagePanelInteractionListener, MouseListener, MouseMotionListener {

        static final int VIEW_MARGIN = 4;
        static final int INTER_PAGE_MARGIN = 10;
        static final int MINSIZE = 100;

        private final boolean vertical;
        private Dimension viewSize, maxPageSize, actualSize;
        private PagePanel[] pagePanels;         // Array of 0..n of panels - usually null
        private Rectangle[] pageRects;          // Array of 0..n of page rectangles, always set
        private Rectangle2D.Float[] fullPageRects;
        private BitSet visiblePages;            // Bitset showing which pages are visible
        private PagePanel mainPagePanel;

        private boolean draggable, pageRectsCentered;
        private Point mouseDownEvent;
        private Point mouseDownScroll;

        View(int orientation) {
            if (orientation != Adjustable.VERTICAL && orientation != Adjustable.HORIZONTAL) {
                throw new IllegalArgumentException("Invalid orientation: "+orientation);
            }
            this.vertical = orientation == Adjustable.VERTICAL;
            setLayout(null); // we will handle layout
            setFocusable(false);
            addMouseListener(this);
            addMouseMotionListener(this);
        }

        /**
         * Called when the document (maybe) has had its pages reordered.
         * Creates "pagePanels" and "visible"
         */
        void reorder() {
            if (pdf == null) {
                if (pagePanels != null) {
                    for (int i=0;i<pagePanels.length;i++) {
                        removePagePanel(i);
                    }
                }
                pagePanels = null;
                pageRects = null;
                visiblePages = null;
                mainPagePanel = null;
                fullPageRects = null;
                viewSize = new Dimension(MINSIZE, MINSIZE);
                maxPageSize = null;
                pageRectsCentered = false;
            } else {
                List pages = pdf.getPages();
                int numpages = pages.size();
                PagePanel[] newPagePanels = new PagePanel[numpages];
                BitSet newVisiblePages = new BitSet();
                if (pagePanels != null) {
                    for (int i=0;i<pagePanels.length;i++) {
                        PagePanel panel = pagePanels[i];
                        if (panel != null) {
                            int index = pages.indexOf(panel.getPage());
                            if (index < 0) {
                                removePagePanel(i);
                            } else {
                                newPagePanels[index] = panel;
                                newVisiblePages.set(index);
                            }
                        }
                    }
                }
                visiblePages = newVisiblePages;
                pagePanels = newPagePanels;

                fullPageRects = new Rectangle2D.Float[numpages];
                for (int i=0;i<fullPageRects.length;i++) {
                    fullPageRects[i] = (Rectangle2D.Float)PagePanel.getFullPageView(pdf.getPage(i));
                    if (fullPageRects[i] == null) {
                        throw new Error();
                    }
                }
            }
            revalidate();
            repaint();
        }

        /**
         * Called when the size of the PagePanels needs to be recalculated.
         */
        void redimension() {
            if (pagePanels == null) {
                reorder();
            }
            if (pdf != null) {
                int numpages = pdf.getNumberOfPages();
                if (fullPageRects == null || fullPageRects.length != numpages) {
                    reorder();
                }
                viewSize = new Dimension(0, 0);
                maxPageSize = new Dimension(0, 0);
                pageRects = new Rectangle[numpages];
                int totwidth = 0, totheight = 0;
                float maxwidth = 0, maxheight = 0;
                int pagex = VIEW_MARGIN;
                int pagey = VIEW_MARGIN;
                pageRectsCentered = false;

                for (int i = 0; i < numpages; i++) {
                    Rectangle2D.Float pageSize = fullPageRects[i];
                    int pagewidth = pointsToPixels(pageSize.width);
                    int pageheight = pointsToPixels(pageSize.height);
                    if (pageSize.width > maxwidth) {
                        maxwidth = pageSize.width;
                    }
                    if (pageSize.height > maxPageSize.height) {
                        maxheight = pageSize.height;
                    }
                    totwidth += pageSize.width;
                    totheight += pageSize.height;
                    Rectangle pageRect = new Rectangle(pagex, pagey, pagewidth, pageheight);
                    pageRects[i] = pageRect;
                    if (vertical) {
                        pagey += pageheight + INTER_PAGE_MARGIN;
                    } else {
                        pagex += pagewidth + INTER_PAGE_MARGIN;
                    }
                }

                if (vertical) { 
                    viewSize.width = pointsToPixels(maxwidth) + VIEW_MARGIN*2;
                    viewSize.height = pointsToPixels(totheight) + VIEW_MARGIN*2 + (numpages-1)*INTER_PAGE_MARGIN; 
                } else {
                    viewSize.width = pointsToPixels(totwidth) + VIEW_MARGIN*2 + (numpages-1)*INTER_PAGE_MARGIN; 
                    viewSize.height = pointsToPixels(maxheight) + VIEW_MARGIN*2;
                }

                int dpi = Util.getScreenResolution(MultiPageDocumentViewport.this);
                maxPageSize = new Dimension(Math.round(maxwidth * dpi / 72), Math.round(maxheight * dpi / 72));
            } else {
                viewSize = new Dimension(MINSIZE, MINSIZE);
                maxPageSize = new Dimension(0, 0);
                pageRects = new Rectangle[0];
            }
            setSize(viewSize);
            setAdjustableValues(getAdjustable(Adjustable.HORIZONTAL).getValue(), getAdjustable(Adjustable.VERTICAL).getValue());
        }

        PagePanel getPagePanel() {
            return mainPagePanel;
        }

        public Dimension getPreferredSize() {
            return viewSize;
        }

        /**
         * Return a BitSet showing which pages should be visible based
         * on the current scrollbar position
         */
        BitSet getVisiblePages() {
            if (pdf == null) {
                return new BitSet();
            }
            Dimension viewport = getPreferredSize();
            Rectangle clip = new Rectangle();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            clip.x = hsb.getValue();
            clip.y = vsb.getValue();
            clip.width = hsb.getVisibleAmount();
            clip.height = vsb.getVisibleAmount();
            BitSet b = new BitSet(pageRects.length);
            for (int i=0;i<pageRects.length;i++) {
                Rectangle r = pageRects[i];
                if (vertical) { 
                    if (r.y < clip.y + clip.height && r.y + r.height > clip.y) {
                        b.set(i);
                    }
                } else {
                    if (r.x < clip.x + clip.width && r.x + r.width > clip.x) {
                        b.set(i);
                    }
                }
            }
            return b;
        }

        Dimension getLargestPageSize() { // NB assumes a zoom of 1.0
            getPreferredSize();
            return maxPageSize;
        }

        public void doLayout() {
            if (pageRects == null) {
                return;
            }
            Rectangle clip = new Rectangle();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            clip.x = hsb.getValue();
            clip.y = vsb.getValue();
            clip.width = hsb.getVisibleAmount();
            clip.height = vsb.getVisibleAmount();
            float dpi = getZoom() * Util.getScreenResolution(MultiPageDocumentViewport.this);

            BitSet newVisiblePages = getVisiblePages();

            if (!getSize().equals(actualSize)) {
                pageRectsCentered = false;
                actualSize = getSize();
            }

            // 1. If page rectangles haven't been centered, do it here and
            // position pagepanels as well.
            if (!pageRectsCentered) {       // If pageRects haven't been centered yet, do it
                Dimension viewport = getPreferredSize();
                int cx = Math.max((actualSize.width - viewport.width) / 2, 0);
                int cy = Math.max((actualSize.height - viewport.height) / 2, 0);
                for (int i=0;i<pageRects.length;i++) {
                    Rectangle rect = pageRects[i];
                    if (vertical) {
                        rect.x = cx + (viewport.width - rect.width)/2;
                    } else {
                        rect.y = cy + (viewport.height - rect.height)/2;
                    }
                    if (pagePanels[i] != null) {
                        pagePanels[i].setBounds(rect);
                    }
                }
                pageRectsCentered = true;
            }

            // 2. Check for changes in the set of visible pages.
            visiblePages.xor(newVisiblePages);
            //            System.out.println("doLayout: VISIBLE="+newVisiblePages+" CHANGE="+visiblePages);
            if (!visiblePages.isEmpty()) {
                for (int i = visiblePages.nextSetBit(0); i >= 0; i = visiblePages.nextSetBit(i+1)) {
                    if (newVisiblePages.get(i)) {     // Page i was added
                        pagePanels[i] = addPagePanel(i, pdf.getPage(i), clip, pageRects[i]);
                    } else {                    // Page i was removed
                        removePagePanel(i);
                    }
                }
            }

            // 3. Of visible pages, position and determine which one is our "main" page.
            // Main page is the first fully visble page, or the page with most area displayed
            int newmainpage = -1, maxvispixels = 0;
            for (int i = newVisiblePages.nextSetBit(0); i >= 0; i = newVisiblePages.nextSetBit(i+1)) {
                repaintPanel(i, null, clip);
                Rectangle visible = clip.intersection(pageRects[i]);
                if (maxvispixels < Integer.MAX_VALUE) {
                    if (visible.width == pageRects[i].width && visible.height == pageRects[i].height) {
                        newmainpage = i;
                        maxvispixels = Integer.MAX_VALUE;
                    }
                    int vispixels = visible.width * visible.height;
                    if (vispixels > maxvispixels) {
                        maxvispixels = vispixels;
                        newmainpage = i;
                    }
                }
            }

            // Maybe fire pageChanged and pagePositionChanged events
            final DocumentPanel docpanel = getDocumentPanel();
            if (docpanel != null && newmainpage >= 0) {
                if (pagePanels[newmainpage] != mainPagePanel) {
                    PDFPage oldPage = (mainPagePanel == null) ? null : mainPagePanel.getPage();
                    mainPagePanel = pagePanels[newmainpage];
                    DocumentPanelEvent dpe = DocumentPanelEvent.createPageChanged(docpanel);
                    dpe.setPreviousPage(oldPage);
                    docpanel.raiseDocumentPanelEvent(dpe);
                }
            }
            visiblePages = newVisiblePages;
        }

        /**
         * @page the page to display
         * @viewClip the clip rectangle
         * @pr the page rectangle
         */
        private PagePanel addPagePanel(int i, PDFPage page, Rectangle viewClip, Rectangle pageRect) {
            PagePanel pagePanel = new PagePanel();
            pagePanel.setViewport(MultiPageDocumentViewport.this);
            pagePanel.setParser(parser);
            if (hints != null) {
                pagePanel.setRenderingHints(hints);
            }
            pagePanel.addPagePanelListener(this);
            pagePanel.addPagePanelInteractionListener(this);
            pagePanel.addMouseListener(this);
            pagePanel.addMouseMotionListener(this);
            pagePanel.setBounds(pageRect);
            pagePanels[i] = pagePanel;
            add(pagePanel);

            pagePanel.raisePagePanelEvent(PagePanelEvent.createPageVisible(pagePanel, page));

            return pagePanel;
        }

        private void repaintPanel(int i, PDFPage page, Rectangle viewClip) {
            if (page == null) {
                page = pagePanels[i].getPage(); // Quickest
                if (page == null) {
                    page = pdf.getPage(i);      // Slowest
                }
            }
            Rectangle pageClip = pageRects[i].intersection(viewClip);
            pageClip.x -= pageRects[i].x;
            pageClip.y -= pageRects[i].y;
            float dpi = getZoom() * Util.getScreenResolution(MultiPageDocumentViewport.this);
            // Next method called on every adjust, but PagePanel filters out unnecessary repaints
            pagePanels[i].setPage(page, fullPageRects[i], dpi, pageClip);
        }

        private void removePagePanel(int i) {
            PagePanel pagePanel = pagePanels[i];
            if (pagePanel != null) {
                pagePanel.removePagePanelListener(this);
                remove(pagePanel);
                pagePanel.dispose();
                pagePanels[i] = null;

                pagePanel.raisePagePanelEvent(PagePanelEvent.createPageHidden(pagePanel, pagePanel.getPage()));
            }
        }

        public void paintComponent(Graphics gg) {
            Graphics2D g = (Graphics2D) gg;
            super.paintComponent(g);
            if (zoom != 0) {
                // Paint page borders
                Color dark = new Color(0xFF666666, false);
                Color medium = new Color(0x80666666, true);
                Color light = new Color(0x30666666, true);
                if (visiblePages != null && pageRects != null) {
                    for (int i = visiblePages.nextSetBit(0); i >= 0; i = visiblePages.nextSetBit(i+1)) {
                        Rectangle r = pageRects[i];
                        int x1 = r.x - 1, y1 = r.y - 1;
                        int x2 = r.x + r.width + 1, y2 = r.y + r.height + 1;
                        g.setColor(dark);
                        g.drawRect(x1, y1, r.width + 2, r.height + 2);
                        y2++;
                        g.setColor(medium);
                        g.drawLine(x1 + 1, y2, x2 - 1, y2);
                        y2++;
                        g.setColor(light);
                        g.drawLine(x1 + 2, y2, x2 - 2, y2);
                    }
                }
            }
        }

        public void pageUpdated(PagePanelEvent event) {
            PagePanelListener[] l;
            synchronized (pagePanelListeners) {
                l = (PagePanelListener[]) pagePanelListeners.toArray(new PagePanelListener[pagePanelListeners.size()]);
            }
            for (int i = 0; i < l.length; i++) {
                l[i].pageUpdated(event);
            }
        }

        public void pageAction(PagePanelInteractionEvent event) {
            PagePanelInteractionListener[] l;
            synchronized (pagePanelInteractionListeners) {
                l = (PagePanelInteractionListener[]) pagePanelInteractionListeners.toArray(new PagePanelInteractionListener[pagePanelInteractionListeners.size()]);
            }
            for (int i = 0; i < l.length; i++) {
                l[i].pageAction(event);
            }
        }


        boolean isDraggable() {
            return draggable;
        }

        void setDraggable(boolean draggable) {
            this.draggable = draggable;
            setCursor(draggable ? CURSOR_GRAB : null);
        }

        public void mouseClicked(MouseEvent event) {}
        public void mouseEntered(MouseEvent event) {}
        public void mouseExited(MouseEvent event) {}
        public void mouseMoved(MouseEvent event) {}

        public void mousePressed(MouseEvent event) {
            if (draggable) {
                smoothScroll(0, 0, null, null);
                Component source = (Component) event.getSource();
                if (source == this) {
                    mouseDownEvent = event.getPoint();
                } else {
                    mouseDownEvent = new Point(event.getX() + source.getX(), event.getY() + source.getY());
                }
                Adjustable sbh = getAdjustable(Adjustable.HORIZONTAL);
                Adjustable sbv = getAdjustable(Adjustable.VERTICAL);
                mouseDownScroll = new Point(sbh.getValue(), sbv.getValue());
                setCursor(CURSOR_GRABBING);
            }
        }

        public void mouseReleased(MouseEvent event) {
            if (draggable) {
                mouseDownEvent = null;
                mouseDownScroll = null;
                setCursor(CURSOR_GRAB);
            }
        }

        public void mouseDragged(MouseEvent event) {
            if (draggable && mouseDownEvent != null) {
                smoothScroll(0, 0, null, null);
                Point point;
                Component source = (Component) event.getSource();
                if (source == this) {
                    point = event.getPoint();
                } else {
                    point = new Point(event.getX() + source.getX(), event.getY() + source.getY());
                }
                int dx = point.x - mouseDownEvent.x;
                int dy = point.y - mouseDownEvent.y;
                Adjustable sbh = getAdjustable(Adjustable.HORIZONTAL);
                Adjustable sbv = getAdjustable(Adjustable.VERTICAL);
                int sbx = sbh.getValue();
                int sby = sbv.getValue();
                setAdjustableValues(mouseDownScroll.x - dx, mouseDownScroll.y - dy);
                int sdx = sbh.getValue() - sbx;
                int sdy = sbv.getValue() - sby;
                mouseDownScroll.x += sdx;
                mouseDownScroll.y += sdy;
            }
        }

        void dispose() {
            if (pagePanels != null) {
                for (int i=0;i<pagePanels.length;i++) {
                    removePagePanel(i);
                }
            }
            pagePanels = null;
            fullPageRects = null;
            pageRects = null;
            visiblePages = null;
        }

        Collection getPagePanels() {
            List l = new ArrayList();
            if (pagePanels != null) {
                for (int i=0;i<pagePanels.length;i++) {
                    if (pagePanels[i] != null) {
                        l.add(pagePanels[i]);
                    }
                }
            }
            return Collections.unmodifiableList(l);
        }

        Rectangle getPageRectangle(int i) {
            if (pageRects == null || i >= pageRects.length) {
                reorder();
            }
            if (pageRects == null || i >= pageRects.length) {
                return null;
            }
            return pageRects[i];
        }
    }

    private class Listener implements ComponentListener, AdjustmentListener {

        private boolean enabled = true;

        void setEnabled(boolean enabled) {
            this.enabled = enabled;
        }

        public void componentResized(ComponentEvent e) {
            int zoommode = getZoomMode();
            switch (zoommode) {
                case ZOOM_FITWIDTH:
                case ZOOM_FITHEIGHT:
                case ZOOM_FIT:
                    setZoom(getTargetZoom(zoommode, null));
                    break;
            }
            smoothScroll(0, 0, null, null);
        }

        public void componentMoved(ComponentEvent e) {
        }

        public void componentShown(ComponentEvent e) {
        }

        public void componentHidden(ComponentEvent e) {
        }

        public void adjustmentValueChanged(AdjustmentEvent e) {
            if (enabled) {
                view.revalidate(); // NB this is required for Windows to actually call doLayout
                view.repaint();
                DocumentPanel docpanel = getDocumentPanel();
                if (docpanel != null && !isSmoothScrolling()) {
                    docpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createPagePositionChanged(docpanel));
                }
            }
        }

    }

}
