// $Id: OpenRecent.java 13935 2011-10-03 10:33:17Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.prefs.Preferences;

/**
 * This feature adds a most-recently-used menu listing the <i>n</i> most
 * recently opened documents. Actuating the menu opens the associated document
 * in the viewer.
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">OpenRecent</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.10.4
 */
public class OpenRecent extends ViewerFeature {

    private int max;
    private Preferences preferences;
    private Opener opener;
    private JMenu openrecentmenu;

    /**
     * Create a new "Open Recent" menu with a default number of items
     */
    public OpenRecent() {
        this(4);
    }

    /**
     * Create a new "Open Recent" menu with the specified number of items
     */
    public OpenRecent(int max) {
        super("OpenRecent");
        this.max = max;
    }

    public void initialize(PDFViewer viewer) {
        super.initialize(viewer);
        preferences = viewer.getPreferences();
        opener = new Opener(viewer);
        for (int i = 0; i < max; i++) {
            String path = null;
            if (preferences!=null) {
                path = preferences.get("features.OpenRecent."+i, null);
                if (path==null) {       // Check old naming convention, and if found migrate it.
                    path = preferences.get("File/mru"+i, null);
                    if (path!=null) {
                        preferences.put("features.OpenRecent."+i, path);
                        preferences.remove("File/mru"+i);
                    }
                }
            }
            if (path != null) {
                File file = new File(path);
                JMenuItem item = viewer.setMenu("File\tOpenRecent\t"+file.getName(), (char)0, false, opener);
                int mask = viewer.getToolkit().getMenuShortcutKeyMask();
                item.setAccelerator(KeyStroke.getKeyStroke((char)('0'+i), mask, false));
                openrecentmenu = (JMenu)((JPopupMenu)item.getParent()).getInvoker();
                item.setActionCommand(path);
            }
        }
        if (openrecentmenu==null) {
            JMenuItem item = viewer.setMenu("File\tOpenRecent\tnull", (char)0, false, null);
            openrecentmenu = (JMenu)((JPopupMenu)item.getParent()).getInvoker();
            openrecentmenu.remove(item);
        }
    }

    /**
     * This method is run by the PDFViewer when the specified file is opened.
     * @param viewer the viewer
     * @param file the File that was opened
     */
    public void opened(PDFViewer viewer, File file) {
        String path = file.getPath();
        // delete existing item that corresponds to this file
        int itemCount = openrecentmenu.getItemCount();
        boolean found = false;
        for (int i=0;!found && i<itemCount;i++) {
            JMenuItem item = openrecentmenu.getItem(i);
            if (item.getActionCommand().equals(path)) {
                openrecentmenu.remove(item);
                openrecentmenu.insert(item, 0);
                found = true;
            }
        }
        if (!found) {
            JMenuItem item = new JMenuItem(file.getName());
            item.setActionCommand(path);
            item.addActionListener(opener);
            openrecentmenu.insert(item, 0);
            // Delete any surplus shortcuts
            while (openrecentmenu.getItemCount() > max) {
                openrecentmenu.remove(max);
            }
            itemCount = openrecentmenu.getItemCount();
        }
        for (int i=0;i<itemCount;i++) {
            JMenuItem item = openrecentmenu.getItem(i);
            int mask = viewer.getToolkit().getMenuShortcutKeyMask();
            item.setAccelerator(KeyStroke.getKeyStroke((char)('0'+i), mask, false));
        }

        if (preferences != null) {
            itemCount = openrecentmenu.getItemCount();
            for (int i = 0; i < itemCount; i++) {
                preferences.put("features.OpenRecent." + i, openrecentmenu.getItem(i).getActionCommand());
            }
        }
    }

    static class Opener implements ActionListener {
        final PDFViewer viewer;

        Opener(PDFViewer viewer) {
            this.viewer = viewer;
        }

        public void actionPerformed(ActionEvent event) {
            String path = event.getActionCommand();
            viewer.loadPDF(new File(path));
        }
    }
}
